"""
Command line tool to generate NXmx NeXus files for Serial Crystallography.
"""

import argparse
import logging
from pathlib import Path
from typing import Tuple

from .. import P, log
from ..beamlines.SSX_chip import CHIP_DICT_DEFAULT
from . import version_parser

logger = logging.getLogger("nexgen.SSX_cli")

usage = "%(prog)s <sub-command> collection-data [options]"
parser = argparse.ArgumentParser(
    usage=usage, description=__doc__, parents=[version_parser]
)
parser.add_argument("--debug", action="store_const", const=True)


def eiger_collection(args):
    logger.info("Create a NeXus file for SSX collection on Eiger.")
    from ..beamlines.SSX_Eiger_nxs import ssx_eiger_writer

    ssx_eiger_writer(
        Path(args.visitpath).expanduser().resolve(),
        args.filename_root,
        args.beamline,
        expt_type=args.expt_type,
        pump_status=args.pump_status,
        num_imgs=args.num_imgs,
        exp_time=args.exp_time,
        det_dist=args.det_dist,
        beam_center=args.beam_center,
        transmission=args.transmission,
        wavelength=args.wavelength,
        start_time=args.start,
        stop_time=args.stop,
        chip_info=CHIP_DICT_DEFAULT,  # TODO This might be better passed as a json/yaml or whatever
        chipmap=args.chipmap,
    )


def tristan_collection(args):
    logger.info("Create a NeXus file for SSX collection on Tristan.")
    from ..beamlines.SSX_Tristan_nxs import ssx_tristan_writer

    ssx_tristan_writer(
        Path(args.visitpath).expanduser().resolve(),
        args.filename_root,
        args.beamline,
        xp_time=args.exp_time,
        det_dist=args.det_dist,
        beam_center=args.beam_center,
        transmission=args.transmission,
        wavelength=args.wavelength,
        start_time=args.start,
        stop_time=args.stop,
        chip_info=CHIP_DICT_DEFAULT,  # TODO This might be better passed as a json/yaml or whatever
        chipmap=args.chipmap,
    )


# Define a parser for the basic collection parameters
class _ImportCollect(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        input_file, visitpath, filename_root = self.find_import_args(values)
        setattr(namespace, self.dest, input_file)
        setattr(namespace, "visitpath", visitpath)
        setattr(namespace, "filename_root", filename_root)

    @staticmethod
    def find_import_args(val) -> Tuple[str]:
        input_file = Path(val).expanduser().resolve()
        visitpath = input_file.parent
        filename_root = P.fullmatch(input_file.stem)[1]
        return input_file, visitpath, filename_root


collect_parser = argparse.ArgumentParser(add_help=False)
# Get visitpath and filename_root out of meta file
collect_parser.add_argument(
    "input_file",
    type=str,
    action=_ImportCollect,
    help="Path to _meta.h5 or _000001.h5 file.",
)
collect_parser.add_argument("beamline", type=str, help="Beamline name.")
collect_parser.add_argument(
    "-det",
    "--det-dist",
    type=float,
    help="Detector distance in mm.",
)
collect_parser.add_argument(
    "-tr",
    "--transmission",
    type=float,
    default=None,
    help="Attenuator transmission.",
)
collect_parser.add_argument(
    "-wl",
    "--wavelength",
    type=float,
    default=None,
    help="Incident beam wavelength, in A.",
)
collect_parser.add_argument(
    "-bc",
    "--beam-center",
    type=float,
    nargs=2,
    default=[0.0, 0.0],
    help="Beam center (x,y) positions.",
)
collect_parser.add_argument(
    "--start", "--start-time", type=str, default=None, help="Collection start time."
)
collect_parser.add_argument(
    "--stop", "--stop-time", type=str, default=None, help="Collection end time."
)

# Define subparsers
subparsers = parser.add_subparsers(
    help="Choose the writer based on the Detector in use on the beamline. \
        Run SSX_nexus <command> --help to see the parameters for each sub-command.",
    required=True,
    dest="sub-command",
)

eiger_parser = subparsers.add_parser(
    "1",
    aliases=["eiger"],
    description=("Trigger Eiger writer."),
    parents=[collect_parser],
)
eiger_parser.add_argument(
    "expt_type",
    type=str,
    choices=["extruder", "fixed-target", "3Dgridscan"],
    help="Choose the type of serial experiment.",
)
eiger_parser.add_argument("num_imgs", type=int, help="Total number of images.")
eiger_parser.add_argument(
    "-e",
    "--exp-time",
    type=float,
    required=True,
    help="Exposure time in s.",
)
eiger_parser.add_argument(
    "-p",
    "--pump-status",
    action="store_true",
    default=False,
    help="Select pump status.",
)
eiger_parser.add_argument("--chipmap", type=str, help="Location of chipmap.")
eiger_parser.set_defaults(func=eiger_collection)

tristan_parser = subparsers.add_parser(
    "2",
    aliases=["tristan"],
    description=("Trigger Tristan writer."),
    parents=[collect_parser],
)
tristan_parser.add_argument(
    "-e",
    "--exp-time",
    type=float,
    required=True,
    help="Total collection time in s.",
)
tristan_parser.add_argument("--chipmap", type=str, help="Location of chipmap.")
tristan_parser.set_defaults(func=tristan_collection)


def main():
    log.config()
    args = parser.parse_args()
    args.func(args)



"""
Tools to read a chip and compute the coordinates of a Serial Crystallography collection.
"""
from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Tuple

# I24 chip tools

CHIP_DICT_DEFAULT = {
    "X_NUM_STEPS": [0, 20],
    "Y_NUM_STEPS": [0, 20],
    "X_STEP_SIZE": [0, 0.125],
    "Y_STEP_SIZE": [0, 0.125],
    "X_START": [0, 0],
    "Y_START": [0, 0],
    "Z_START": [0, 0],
    "X_NUM_BLOCKS": [0, 8],
    "Y_NUM_BLOCKS": [0, 8],
    "X_BLOCK_SIZE": [0, 3.175],
    "Y_BLOCK_SIZE": [0, 3.175],
    "N_EXPOSURES": [0, 1],
    "PUMP_REPEAT": [0, 0],
}


@dataclass
class Chip:
    """
    Define a fixed target chip.

    Args:
        name (str): Description of the chip.
        num_steps (List[int] | Tuple[int]): Number of windows in each block.
        step_size (List[float] | Tuple[float]): Size of each window (distance between the centers in x and y direction).
        num_blocks (List[int] | Tuple[int]): Total number of blocks in the chip.
        block_size (List[int] | Tuple[int]): Size of each block.
        start_pos (List[float]): Start coordinates (x,y,z)
    """

    name: str

    num_steps: List[int, int] | Tuple[int, int]
    step_size: List[float, float] | Tuple[float, float]
    num_blocks: List[int, int] | Tuple[int, int]
    block_size: List[float, float] | Tuple[float, float]

    start_pos: List[float, float, float] = field(default_factory=[0.0, 0.0, 0.0])

    def tot_blocks(self) -> int:
        return self.num_blocks[0] * self.num_blocks[1]

    def tot_windows_per_block(self) -> int:
        return self.num_steps[0] * self.num_steps[1]

    def window_size(self) -> Tuple[float]:
        return (
            self.num_steps[0] * self.step_size[0],
            self.num_steps[1] * self.step_size[1],
        )

    def chip_size(self) -> Tuple[float]:
        return (
            self.num_blocks[0] * self.block_size[0],
            self.num_blocks[1] * self.block_size[1],
        )


def fullchip_conversion_table(chip: Chip) -> Dict:
    coords = []
    table = {}
    for i in range(chip.num_blocks[0]):
        if i % 2 == 0:
            for j in range(chip.num_blocks[1]):
                coords.append((i, j))
        else:
            for j in range(chip.num_blocks[1] - 1, -1, -1):
                coords.append((i, j))
    for k, v in zip(range(1, chip.tot_blocks() + 1), coords):
        table[f"%0{2}d" % k] = v
    return table


def read_chip_map(mapfile: Path | str, x_blocks: int, y_blocks: int) -> Dict:
    """
    Read the .map file for the current collection on a chip.

    Args:
        mapfile (Path | str): Path to .map file. If None, assumes fullchip.
        x_blocks (int): Total number of blocks in x direction in the chip.
        y_blocks (int): Total number of blocks in y direction in the chip.

    Returns:
        Dict: A dictionary whose values indicate either the coordinates on the chip \
            of the scanned blocks, or a string indicating that the whole chip is being scanned.
    """
    if mapfile is None:
        # Assume it's a full chip
        return {"all": "fullchip"}

    with open(mapfile, "r") as f:
        chipmap = f.read()

    block_list = []
    max_num_blocks = x_blocks * y_blocks
    for n, line in enumerate(chipmap.rsplit("\n")):
        if n == max_num_blocks:
            break
        k = line[:2]
        v = line[-1:]
        if v == "1":
            block_list.append(k)
    if len(block_list) == max_num_blocks:
        # blocks["fullchip"] = len(block_list)
        return {"all": "fullchip"}

    blocks = {}
    for b in block_list:
        x = int(b) // x_blocks if int(b) % x_blocks != 0 else (int(b) // x_blocks) - 1
        if x % 2 == 0:
            val = x * x_blocks + 1
            y = int(b) - val
        else:
            val = (x + 1) * x_blocks
            y = val - int(b)
        blocks[b] = (x, y)
    return blocks


def fullchip_blocks_conversion(blocks: Dict[Tuple, List], chip: Chip) -> Dict:
    new_blocks = {}
    table = fullchip_conversion_table(chip)
    for kt, vt in table.items():
        for kb, vb in blocks.items():
            if kb == vt:
                new_blocks[kt] = vb
    return new_blocks


def compute_goniometer(
    chip: Chip,
    axes: List,
    blocks: Dict = None,
    full: bool = False,
    ax0: str = "sam_x",
    ax1: str = "sam_y",
) -> Tuple[Dict]:
    """Compute the start and end coordinates of a chip scan.

    The function returns two dictionaries - one for start and one for end positions - associating a list of axes values to each scanned block.
    All values corresponding to axes unrelated to the chip will be set automatically to 0.
    If full is True, the blocks argument will be overridden and coordinates will be calculated for every block in the chip.

    Args:
        chip (Chip): General description of the chip schematics: number and size of blocks, size and step of each window, start positions.
        axes (List): List of all the goniometer axes.
        blocks (Dict, optional): Scanned blocks. Defaults to None.
        full (bool, optional): True if all blocks have been scanned. Defaults to False.
        ax0 (str, optional): Goniometer axis corresponding to 'x' on chip. Defaults to "sam_x".
        ax1 (str, optional): Goniometer axis corresponding to 'y' on chip. Defaults to "sam_y".

    Returns:
        start, end (Tuple[Dict]): Goniometer start and end coordinates for each block.

    Raises:
        ValueError: If one or both of the axes names passed as input are not in the list og goniometer axes.
    """
    x0 = chip.start_pos[0]
    y0 = chip.start_pos[1]

    if ax0 not in axes or ax1 not in axes:
        raise ValueError(
            "Axis not found in the list of goniometer axes. Please check your input."
            f"Goniometer axes: {axes}. Looking for {ax0} and {ax1}."
        )

    num_axes = len(axes)
    idx_X = axes.index(ax0)
    idx_Y = axes.index(ax1)

    starts = {}
    ends = {}

    if full is True:
        for x in range(chip.num_blocks[0]):
            x_start = x0 + x * chip.block_size[0]
            if x % 2 == 0:
                for y in range(chip.num_blocks[1]):
                    y_start = y0 + y * chip.block_size[1]
                    x_end = x_start + chip.num_steps[0] * chip.step_size[0]
                    y_end = y_start + chip.num_steps[1] * chip.step_size[1]
                    starts[(x, y)] = [
                        x_start if i == idx_X else y_start if i == idx_Y else 0.0
                        for i in range(num_axes)
                    ]
                    ends[(x, y)] = [
                        x_end if i == idx_X else y_end if i == idx_Y else 0.0
                        for i in range(num_axes)
                    ]
            else:
                for y in range(chip.num_blocks[1] - 1, -1, -1):
                    y_end = y0 + y * chip.block_size[1]
                    x_end = x_start + chip.num_steps[0] * chip.step_size[0]
                    y_start = y_end + (chip.num_steps[1] - 1) * chip.step_size[1]
                    starts[(x, y)] = [
                        x_start if i == idx_X else y_start if i == idx_Y else 0.0
                        for i in range(num_axes)
                    ]
                    ends[(x, y)] = [
                        x_end if i == idx_X else y_end if i == idx_Y else 0.0
                        for i in range(num_axes)
                    ]
    else:
        for k, v in blocks.items():
            x_start = x0 + v[0] * chip.block_size[0]
            if v[0] % 2 == 0:
                y_start = x0 + v[1] * chip.block_size[1]
                x_end = x_start + chip.num_steps[0] * chip.step_size[0]
                y_end = y_start + chip.num_steps[1] * chip.step_size[1]
            else:
                y_end = x0 + v[1] * chip.block_size[1]
                x_end = x_start + chip.num_steps[0] * chip.step_size[0]
                y_start = y_end + (chip.num_steps[1] - 1) * chip.step_size[1]
            starts[k] = [
                round(x_start, 3)
                if i == idx_X
                else round(y_start, 3)
                if i == idx_Y
                else 0.0
                for i in range(num_axes)
            ]
            ends[k] = [
                round(x_end, 3)
                if i == idx_X
                else round(y_end, 3)
                if i == idx_Y
                else 0.0
                for i in range(num_axes)
            ]

    return starts, ends



"""
Writer for NeXus format files.
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import Any, Dict, List, Tuple

import h5py
import numpy as np
from numpy.typing import ArrayLike

from .. import get_filename_template, reframe_arrays, units_of_time
from ..tools.DataWriter import generate_event_files, generate_image_files
from ..tools.MetaReader import overwrite_beam, overwrite_detector
from ..tools.VDS_tools import image_vds_writer, vds_file_writer
from . import (
    calculate_scan_range,
    find_grid_scan_axes,
    find_number_of_images,
    find_osc_axis,
)
from .NXclassWriters import (
    write_NXdata,
    write_NXdatetime,
    write_NXdetector,
    write_NXdetector_module,
    write_NXentry,
    write_NXinstrument,
    write_NXnote,
    write_NXsample,
    write_NXsource,
)


# Define the scan.
def ScanReader(
    goniometer: Dict,
    data_type: str = "images",
    n_images: int | Tuple = None,
    snaked: bool = True,
    osc_axis: str = None,
) -> Tuple[Dict, Dict]:
    """
    Read the information passed from the goniometer and return a definition of the scan.

    Args:
        goniometer (Dict): Goniometer geometry definition.
        data_type (str, optional): Type of data being written, can be images of events. Defaults to "images".
        n_images (int | Tuple, optional): Total number of images to write. If passed, \
                                    the number of images will override the axis_increment value of the rotation scan. \
                                    Defaults to None.
        snaked (bool, optional): 2D scan parameter. If True, defines a snaked grid scan. Defaults to True.

    Raises:
        ValueError: If the total number of images passed doesn't match the number of scan points when dealing with a 2D/3D scan.

    Returns:
        Tuple[Dict, Dict]: Two separate dictionaries. The first defines the rotation scan, the second the linear/grid scan. \
                            When dealing with a set of stills or a simple rotation scan, the second value will return None.
    """
    logger = logging.getLogger("nexgen.ScanReader")
    logger.setLevel(logging.DEBUG)
    # First find which axes deifne a rotation/translation scan
    if osc_axis is None:
        osc_axis = find_osc_axis(
            goniometer["axes"],
            goniometer["starts"],
            goniometer["ends"],
            goniometer["types"],
        )
    osc_idx = goniometer["axes"].index(osc_axis)
    transl_axes = find_grid_scan_axes(
        goniometer["axes"],
        goniometer["starts"],
        goniometer["ends"],
        goniometer["types"],
    )

    # If there's a linear/grid scan, get dictionary
    if len(transl_axes) > 0:
        transl_idx = [goniometer["axes"].index(j) for j in transl_axes]
        transl_start = [goniometer["starts"][i] for i in transl_idx]
        transl_end = [goniometer["ends"][i] for i in transl_idx]
        transl_increment = [goniometer["increments"][i] for i in transl_idx]
        if n_images and type(n_images) is int:
            TRANSL = calculate_scan_range(
                transl_axes,
                transl_start,
                transl_end,
                transl_increment,
                # (n_images,),
                snaked=snaked,
            )
        elif n_images and type(n_images) is tuple:
            TRANSL = calculate_scan_range(
                transl_axes, transl_start, transl_end, n_images=n_images, snaked=snaked
            )
        else:
            TRANSL = calculate_scan_range(
                transl_axes, transl_start, transl_end, transl_increment, snaked=snaked
            )
        logger.debug(f"{len(transl_axes)} scan axis/axes found (translation).")
    else:
        TRANSL = None

    # Once that's defined, go through the various cases
    # Return either 2 dictionaries or (Dict, None)
    if data_type == "events" and len(transl_axes) == 0:
        OSC = {osc_axis: (goniometer["starts"][osc_idx], goniometer["ends"][osc_idx])}
    elif data_type == "events" and len(transl_axes) > 0:
        OSC = {osc_axis: (goniometer["starts"][osc_idx], goniometer["ends"][osc_idx])}
        # Overwrite TRANSL
        for k, s, e in zip(transl_axes, transl_start, transl_end):
            TRANSL[k] = (s, e)
    else:
        if n_images is None and len(transl_axes) == 0:
            OSC = calculate_scan_range(
                [osc_axis],
                [goniometer["starts"][osc_idx]],
                [goniometer["ends"][osc_idx]],
                axes_increments=[goniometer["increments"][osc_idx]],
                rotation=True,
            )
        elif n_images is None and len(transl_axes) > 0:
            ax = transl_axes[0]
            n_images = len(TRANSL[ax])
            OSC = calculate_scan_range(
                [osc_axis],
                [goniometer["starts"][osc_idx]],
                [goniometer["ends"][osc_idx]],
                n_images=n_images,
                rotation=True,
            )
        elif n_images is not None and len(transl_axes) > 0:
            ax = transl_axes[0]
            n_images = np.prod(n_images) if type(n_images) is tuple else n_images
            if n_images != len(TRANSL[ax]):
                raise ValueError(
                    "The value passed as the total number of images doesn't match the number of scan points, please check the input."
                )
            #
            OSC = calculate_scan_range(
                [osc_axis],
                [goniometer["starts"][osc_idx]],
                [goniometer["ends"][osc_idx]],
                n_images=n_images,
                rotation=True,
            )
        else:
            n_images = np.prod(n_images) if type(n_images) is tuple else n_images
            OSC = calculate_scan_range(
                [osc_axis],
                [goniometer["starts"][osc_idx]],
                [goniometer["ends"][osc_idx]],
                n_images=n_images,
                rotation=True,
            )

    logger.debug(f"{osc_axis} set as rotation axis.")
    return OSC, TRANSL


# Write NeXus base classes
def call_writers(
    nxsfile: h5py.File,
    datafiles: List[Path | str],
    coordinate_frame: str,
    data_type: Tuple[str, int],
    goniometer: Dict[str, Any],
    detector: Dict[str, Any],
    module: Dict[str, Any],
    source: Dict[str, Any],
    beam: Dict[str, Any],
    attenuator: Dict[str, Any],
    osc_scan: Dict[str, ArrayLike],
    transl_scan: Dict[str, ArrayLike] = None,
    metafile: Path | str = None,
    link_list: List = None,
    sample_depends_on: str = None,
):
    """
    Call the writers for the NeXus base classes.

    Args:
        nxsfile (h5py.File): NeXus file to be written.
        datafiles (List[Path |str]): List of at least 1 Path object to a HDF5 data file.
        coordinate_frame (str): Coordinate system being used. Accepted frames are imgcif and mcstas.
        data_type (Tuple[str, int]): Images or event-mode data, and eventually how many are being written.
        goniometer (Dict[str, Any] Goniometer geometry description.
        detector (Dict[str, Any]): Detector specific parameters and its axes.
        module (Dict[str, Any]): Geometry and description of detector module.
        source (Dict[str, Any]): Facility information.
        beam (Dict[str, Any]): Beam properties.
        attenuator (Dict[str, Any]): Attenuator properties.
        osc_scan (Dict[str, ArrayLike]): Axis defining the rotation scan. It should be passed even when still.
        transl_scan (Dict[str, ArrayLike], optional): Axes defining a linear or 2D scan. Defaults to None.
        metafile (Path | str, optional): File containing the metadata. Defaults to None.
        link_list (List, optional): List of datasets that can be copied from the metafile. Defaults to None.
        sample_depends_on (str, optional): Axis on which the sample depends on. Defaults to None.
    """
    logger = logging.getLogger("nexgen.Call")
    logger.setLevel(logging.DEBUG)
    logger.info("Calling the writers ...")

    # Split vectors and offsets in goniometer and detector for writing
    reframe_arrays(
        goniometer,
        detector,
        module,
        coordinate_frame,
    )

    # Check that filenames are paths
    if all(isinstance(f, Path) for f in datafiles) is False:
        datafiles = [Path(f).expanduser().resolve() for f in datafiles]

    if metafile:
        if type(metafile) is str:
            metafile = Path(metafile).expanduser().resolve()

    # NXdata: entry/data
    write_NXdata(
        nxsfile,
        datafiles,
        goniometer,
        data_type,
        osc_scan,
        transl_scan,
    )

    # NXinstrument: entry/instrument
    write_NXinstrument(
        nxsfile,
        beam,
        attenuator,
        source,
    )

    # NXdetector: entry/instrument/detector
    write_NXdetector(
        nxsfile,
        detector,
        data_type,
        metafile,
        link_list,
    )

    # NXmodule: entry/instrument/detector/module
    write_NXdetector_module(
        nxsfile,
        module,
        detector["image_size"],  # [::-1],
        detector["pixel_size"],
        beam_center=detector["beam_center"],
    )

    # NXsource: entry/source
    write_NXsource(nxsfile, source)

    # NXsample: entry/sample
    write_NXsample(
        nxsfile,
        goniometer,
        data_type,
        osc_scan,
        transl_scan,
        sample_depends_on,
    )


# Write NeXus directly from scope_extract
def write_nexus_from_scope(
    nxs_file: h5py.File,
    goniometer,
    detector,
    module,
    beam,
    attenuator,
    source,
    coordinate_frame: str = "mcstas",
    datafiles: List[Path] = None,
    data_type: Tuple[str, int] = None,
    **params,
):
    """
    Write a new NXmx NeXus file taking scope extracts as input.

    This function writes either a NeXus file for an existing dataset or a demo NeXus file with a blank dataset using the information passed from phil parameters.

    Args:
        nxs_file (h5py.File): NeXus file handle
        goniometer (freephil.common.scope_extract): Scope extract defining the goniometer geometry.
        detector (freephil.common.scope_extract): Scope extract defining the detector and its axes.
        module (freephil.common.scope_extract): Scope extract defining defining geometry and description of module.
        beam (freephil.common.scope_extract): Scope extract defining properties of beam.
        attenuator (freephil.common.scope_extract): Scope extract defining transmission.
        source (freephil.common.scope_extract): Scope extract describing the facility.
        coordinate_frame (str, optional): Coordinate system used to define the vectors. Defaults to "mcstas".
        datafiles (List[Path], optional): HDF5 data files from an existing collection. Defaults to None.
        data_type (Tuple[str, int], optional): Description of the data and eventually how many images make up the dataset. Defaults to None.

    Keyword arguments (**params):
        meta (Tuple[Path | str, List | None]): Metafile information, path to _meta.h5 file and eventual list of values not to look for in it.
        vds (str): If passed, a Virtual DataSet will be written. Accepted values: "dataset", "file".
        tristanSpec (freephil.common.scope_extract): Scope extract defining properties of a Tristan detector.
        timestamps (Tuple[datetime.datetime | None]): Start and end time of the collection, if known. Passed as datetime.datetime.
        notes (Tuple[str, Dict]): Any eventual notes that would be worth recording in the NeXus file.

    Raises:
        ValueError: If data_type has not been passed when writing blank data.
    """
    writer_logger = logging.getLogger("nexgen.WritefromScope")
    writer_logger.setLevel(logging.DEBUG)

    writer_logger.info("Write NXmx NeXus file from scope extract input.")

    # Check if it's a NeXus file for an existing collection or a new blank dataset.
    # And define data_type
    if datafiles:
        writer_logger.info("Writing a NeXus file for an existing collection")
        if detector.mode == "events":
            data_type = ("events", None)
            writer_logger.info("Data type: events.")
        else:
            # Find total number of images that have been written across the files.
            if len(datafiles) == 1:
                with h5py.File(datafiles[0], "r") as f:
                    num_images = f["data"].shape[0]
            else:
                num_images = find_number_of_images(datafiles)
            data_type = ("images", num_images)
            writer_logger.info(
                "Data type: images. \n" f"Total number of images: {num_images}"
            )
    else:
        writer_logger.warning("No data files have been passed.")
        writer_logger.info("Writing a demo NeXus file with blank data.")
        if data_type is None:
            writer_logger.error(
                "When writing a demo with blank data, a data_type tuple should be passed."
                "data_type is a Tuple[str, int | None],"
                "It should cointain a string indicating whether images or events are being written and eventually the number of images."
            )
            raise ValueError(
                "Missing data type information."
                "Please pass the type of data being written ('images'/'events') and eventually the number of images."
            )

    # Log goniometer information
    writer_logger.info("Goniometer information")
    for j in reversed(range(len(goniometer.axes))):
        vector = goniometer.vectors[3 * j : 3 * j + 3]
        writer_logger.info(
            f"Goniometer axis: {goniometer.axes[j]} => {vector} ({goniometer.types[j]}) on {goniometer.depends[j]}. {goniometer.starts[j]} {goniometer.ends[j]} {goniometer.increments[j]}"
        )
    writer_logger.info("\n")

    # Define rotation and translation axes
    OSC, TRANSL = ScanReader(
        goniometer.__dict__,
        data_type[0],
        n_images=data_type[1],
        snaked=params["snaked"] if "snaked" in params.keys() else True,
    )

    # Log scan information
    writer_logger.info("Coordinate system: %s" % coordinate_frame)
    writer_logger.info(f"Rotation scan axis: {list(OSC.keys())[0]}.")
    writer_logger.info(
        f"Scan from {list(OSC.values())[0][0]} to {list(OSC.values())[0][-1]}.\n"
    )
    if TRANSL:
        writer_logger.info(f"Scan along the {list(TRANSL.keys())} axes.")
        for k, v in TRANSL.items():
            writer_logger.info(f"{k} scan from {v[0]} to {v[-1]}.")
    writer_logger.info("\n")

    # Fix the number of images if not already there
    if data_type[0] == "images" and data_type[1] is None:
        data_type = ("images", len(list(OSC.values())[0]))
        writer_logger.warning(f"Total number of images updated to: {data_type[1]} \n")

    # If writing blank data, figure out how many files will need to be written
    if not datafiles:
        writer_logger.info("Calculating number of files to write ...")
        if data_type[0] == "events":
            # Determine the number of files. Write one file per module.
            n_files = 10 if "10M" in detector.description.upper() else 2
        else:
            # The maximum number of images being written each dataset is 1000
            if data_type[1] <= 1000:
                n_files = 1
            else:
                n_files = int(np.ceil(data_type[1] / 1000))

        writer_logger.info(
            f"{n_files} file(s) containing blank {data_type[0]} data to be written."
        )

        # Get datafile list
        datafile_template = get_filename_template(
            Path(nxs_file.filename).expanduser().resolve()
        )
        datafiles = [
            Path(datafile_template % (n + 1)).expanduser().resolve()
            for n in range(n_files)
        ]

        writer_logger.info("Calling data writer ...")
        # Write data files
        if data_type[0] == "images":
            generate_image_files(
                datafiles, detector.image_size, detector.description, data_type[1]
            )
        else:
            exp_time = units_of_time(detector.exposure_time)
            generate_event_files(
                datafiles, data_type[1], detector.description, exp_time.magnitude
            )

    # If _meta.h5 file is passed, look through it for relevant information
    if "meta" in params.keys():
        writer_logger.info("Looking through _meta.h5 file for metadata.")
        metafile = (
            Path(params["meta"]).expanduser().resolve()
            if isinstance(params["meta"], str) is False
            else params["meta"]
        )
        do_not_link = params["ignore_meta"]
        # overwrite detector, overwrite beam, get list of links for nxdetector and nxcollection
        with h5py.File(metafile, "r") as mf:
            overwrite_beam(mf, detector.description, beam)
            link_list = overwrite_detector(mf, detector, do_not_link)
    else:
        writer_logger.info("No _meta file has been passed.")
        metafile = None
        link_list = None

    # Log detector information
    if "TRISTAN" in detector.description.upper() and "tristanSpec" in params.keys():
        from ..command_line import add_tristan_spec

        add_tristan_spec(detector, params["tristanSpec"])

    writer_logger.info("Detector information")
    writer_logger.info(f"{detector.description}")
    writer_logger.info(
        f"Sensor made of {detector.sensor_material} x {detector.sensor_thickness}"
    )
    if data_type[0] == "images":
        writer_logger.info(
            f"Trusted pixels > {detector.underload} and < {detector.overload}"
        )
    writer_logger.info(
        f"Image is a {detector.image_size} array of {detector.pixel_size} pixels"
    )

    writer_logger.info("Detector axes:")
    for j in range(len(detector.axes)):
        vector = detector.vectors[3 * j : 3 * j + 3]
        writer_logger.info(
            f"Detector axis: {detector.axes[j]} => {vector} ({detector.types[j]}) on {detector.depends[j]}. {detector.starts[j]}"
        )

    if detector.flatfield is None:
        writer_logger.info("No flatfield applied")
    else:
        writer_logger.info(f"Flatfield correction data lives here {detector.flatfield}")

    if detector.pixel_mask is None:
        writer_logger.info("No bad pixel mask for this detector")
    else:
        writer_logger.info(f"Bad pixel mask lives here {detector.pixel_mask}")

    writer_logger.info("Module information")
    writer_logger.warning(f"module_offset field setting: {module.module_offset}")
    writer_logger.info(f"Number of modules: {module.num_modules}")
    writer_logger.info(f"Fast axis at datum position: {module.fast_axis}")
    writer_logger.info(f"Slow_axis at datum position: {module.slow_axis} \n")

    # Log source information
    writer_logger.info("Source information")
    writer_logger.info(f"Facility: {source.name} - {source.type}.")
    writer_logger.info(f"Beamline: {source.beamline_name} \n")

    # Start writing NeXus file
    write_NXentry(nxs_file)

    # Call the writers
    call_writers(
        nxs_file,
        datafiles,
        coordinate_frame,
        data_type,
        goniometer.__dict__,
        detector.__dict__,
        module.__dict__,
        source.__dict__,
        beam.__dict__,
        attenuator.__dict__,
        OSC,
        TRANSL,
        metafile,
        link_list,
    )

    # Write VDS if prompted
    if "vds" in params.keys():
        if data_type[0] == "images" and params["vds"] == "dataset":
            writer_logger.info(
                "Calling VDS writer to write a Virtual Dataset under /entry/data/data"
            )
            image_vds_writer(nxs_file, (data_type[1], *detector.image_size))
        elif data_type[0] == "images" and params["vds"] == "file":
            writer_logger.info(
                "Calling VDS writer to write a Virtual Dataset file and relative link."
            )
            vds_file_writer(nxs_file, datafiles, (data_type[1], *detector.image_size))
        else:
            writer_logger.info("VDS won't be written.")

    # Write timestamps if prompted
    if "timestamps" in params.keys():
        timestamps = params["timestamps"]
        writer_logger.info("Writing recorded timestamps.")
        # NX_DATE_TIME: /entry/start_time and /entry/end_time
        if timestamps[0] is not None:
            write_NXdatetime(nxs_file, (timestamps[0], None))
        if timestamps[1] is not None:
            write_NXdatetime(nxs_file, (None, timestamps[1]))

    # Write any notes that might have been passed as NXnote
    if "notes" in params.keys():
        writer_logger.info(
            f"Writing NXnote in requested location {params['notes'][0]}."
        )
        write_NXnote(nxs_file, params["notes"][0], params["notes"][1])



"""
General tools to copy metadata from NeXus files.
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import List

import h5py

from .. import get_nexus_filename
from ..nxs_write import create_attributes
from . import get_nexus_tree

copy_logger = logging.getLogger("nexgen.CopyNeXus")


def images_nexus(
    data_file: List[Path | str],
    original_nexus: Path | str,
    simple_copy: bool = True,
    skip_group: List[str] = ["NXdata"],
) -> str:
    """
    Copy NeXus metadata for images.

    Args:
        data_file (List[Path  |  str]): HDF5 file with images.
        original_nexus (Path  |  str): Original NeXus file with experiment metadata.
        simple_copy (bool, optional): Copy everything from the original NeXus file. Defaults to True.
        skip_group (List[str], optional): If simple_copy is False, list of NX_class objects to skip when copying.
                                        Defaults to ["NXdata"].

    Returns:
        nxs_filename (str): Filename of new NeXus file.
    """
    copy_logger.info("Copying NeXus file for image dataset ...")
    data_file = [Path(d).expanduser().resolve() for d in data_file]
    original_nexus = Path(original_nexus).expanduser().resolve()
    nxs_filename = get_nexus_filename(data_file[0], copy=True)
    copy_logger.info(f"New NeXus file name: {nxs_filename}")
    with h5py.File(original_nexus, "r") as nxs_in, h5py.File(
        nxs_filename, "x"
    ) as nxs_out:
        if simple_copy is True:
            # Copy the whole tree
            get_nexus_tree(nxs_in, nxs_out, skip=False)
        else:
            copy_logger.warning(
                f"The following NX base classes will not be copied: {skip_group}"
            )
            nxs_out.attrs["default"] = "entry"
            # Copy the whole tree except for nxdata and whatever other group was passed.
            nxentry = get_nexus_tree(nxs_in, nxs_out, skip=True, skip_obj=skip_group)
            copy_logger.info(f"Re write NXdata with link to {data_file[0]}.")
            if "NXdata" in skip_group:  # it always is...
                # Create nxdata group
                nxdata = nxentry.create_group("data")
                # Find and copy only scan axis information
                for k in nxs_in["entry/data"].keys():
                    if "depends_on" in nxs_in["entry/data"][k].attrs.keys():
                        ax = k
                        nxs_in["entry/data"].copy(k, nxdata)
                create_attributes(
                    nxdata, ("NX_class", "axes", "signal"), ("NXdata", ax, "data")
                )
                # Add link to data
                if len(data_file) == 1:
                    nxdata["data"] = h5py.ExternalLink(data_file[0].name, "data")
                else:
                    for filename in data_file:
                        nxdata[filename.stem] = h5py.ExternalLink(filename.name, "data")
    return nxs_filename.as_posix()


def pseudo_events_nexus(
    data_file: List[Path | str],
    original_nexus: Path | str,
) -> str:
    """
    Copy NeXus metadata for pseudo event mode data.

    Args:
        data_file (List[Path  |  str]): HDF5 with pseud event data.
        original_nexus (Path  |  str): Original NeXus file with experiment metadata.

    Returns:
        nxs_filename (str): Filename of new NeXus file.
    """
    copy_logger.info("Copying NeXus file for events dataset ...")
    data_file = [Path(d).expanduser().resolve() for d in data_file]
    original_nexus = Path(original_nexus).expanduser().resolve()
    nxs_filename = get_nexus_filename(data_file[0], copy=True)
    copy_logger.info(f"New NeXus file name: {nxs_filename}")
    with h5py.File(original_nexus, "r") as nxs_in, h5py.File(
        nxs_filename, "x"
    ) as nxs_out:
        nxs_out.attrs["default"] = "entry"
        # Copy the whole tree except for nxdata
        nxentry = get_nexus_tree(nxs_in, nxs_out)
        # Create nxdata group
        nxdata = nxentry.create_group("data")
        # Find and copy only scan axis information
        for k in nxs_in["entry/data"].keys():
            if "depends_on" in nxs_in["entry/data"][k].attrs.keys():
                ax = k
                nxs_in["entry/data"].copy(k, nxdata)
        create_attributes(
            nxdata, ("NX_class", "axes", "signal"), ("NXdata", ax, "data")
        )
        # Add link to data
        if len(data_file) == 1:
            with h5py.File(data_file[0], "r") as fout:
                for k in fout.keys():
                    nxdata[k] = h5py.ExternalLink(fout.filename, k)
        else:
            for filename in data_file:
                nxdata[filename.stem] = h5py.ExternalLink(filename.name, "/")
    return nxs_filename.as_posix()



"""
Command line tool to copy experiment metadata from one NeXus file to the other.
"""

import argparse
import logging
import sys
from pathlib import Path

import freephil

from .. import log
from ..nxs_copy import CopyNexus, CopyTristanNexus
from . import config_parser, full_copy_parser, tristan_copy_parser, version_parser

# Define a logger object and a formatter
logger = logging.getLogger("nexgen.CopyNeXus")

# Phil scopes
general_scope = freephil.parse(
    """
    input{
      original_nexus = None
        .type = path
        .help = "NeXus file to be copied."
      data_filename = None
        .multiple = True
        .type = path
        .help = "HDF5 data file."
      data_type = *images events
        .type = choice
        .help = "Type of data in the HDF5 file, can be either images or events."
      simple_copy = False
        .type = bool
        .help = "If True, the full NeXus tree is copied."
      skip = NXdata
        .multiple = True
        .optional = False
        .type = str
        .help = "NX_class object, or list of, to be skipped when copying metadata.
                If called, it will always first skip NXdata."
    }
    """
)

tristan_scope = freephil.parse(
    """
    input {
      tristan_nexus = None
        .type = path
        .help = "NeXus file associated with Tristan detector"
      data_filename = None
        .multiple = True
        .type = path
        .help = "HDF5 file with binned images"
      experiment_type = stationary *rotation
        .type = choice
        .help = "Specify whether an experiment is stationary or a rotation scan. Defaults to rotation."
      write_mode = r+ w *x a
        .type = choice
        .help = "Specify write mode for new NeXus file."
    }
    """
)

# Parse command line arguments
parser = argparse.ArgumentParser(
    description="Copy metadata from input NeXus file.",
    parents=[version_parser],
)

parser.add_argument("--debug", action="store_const", const=True)


# CLIs
def copy_nexus(args):
    clai = general_scope.command_line_argument_interpreter()
    working_phil = general_scope.fetch(clai.process_and_fetch(args.phil_args))
    params = working_phil.extract()

    if args.show_config:
        working_phil.show(attributes_level=args.attributes_level)
        sys.exit()

    # Configure logging
    log.config()

    logger.info("Copy metadata from one NeXus file to another.")

    # Path to data file and original nexus file
    data_file = [Path(d).expanduser().resolve() for d in params.input.data_filename]
    nexus_file = Path(params.input.original_nexus).expanduser().resolve()
    logger.info(f"NeXus file to be copied: {nexus_file}")
    logger.info(f"Input data to be saved in NeXus file: {data_file}")

    logger.info(f"Data type: {params.input.data_type}")
    if params.input.simple_copy is True:
        logger.info(f"{nexus_file} will be copied in its entirety.")
    else:
        logger.info(
            f"The following groups will not be copied from NXentry of {nexus_file}: {params.input.skip}"
        )
    try:
        if params.input.data_type == "images":
            new_nxs = CopyNexus.images_nexus(
                data_file,
                nexus_file,
                simple_copy=params.input.simple_copy,
                skip_group=params.input.skip,
            )
        elif params.input.data_type == "events":
            new_nxs = CopyNexus.pseudo_events_nexus(
                data_file,
                nexus_file,
            )
        logger.info(f"File {nexus_file} correctly copied to {new_nxs}.")
    except Exception as err:
        logger.info(f"File {nexus_file} could not be copied.")
        logger.exception(err)


def copy_tristan_nexus(args):
    clai = tristan_scope.command_line_argument_interpreter()
    working_phil = tristan_scope.fetch(clai.process_and_fetch(args.phil_args))
    params = working_phil.extract()

    if args.show_config:
        working_phil.show(attributes_level=args.attributes_level)
        sys.exit()

    # Configure logging
    log.config()

    logger.info("Copy metadata from Tristan NeXus file.")

    # Path to data and original nexus file
    data_file = [Path(d).expanduser().resolve() for d in params.input.data_filename]
    nexus_file = Path(params.input.tristan_nexus).expanduser().resolve()
    logger.info(f"Working directory: {data_file[0].parent}")
    logger.info(f"NeXus file to be copied: {nexus_file}")
    logger.info(f"Input data to be saved in NeXus file: {data_file}")

    try:
        if params.input.experiment_type == "stationary":
            logger.info(
                "Copying metadata for a stationary dataset. \n"
                "This means either a single image or a pump-probe experiment.\n"
                "The 'scan_axis' will be a single scalar."
            )
            nxs_img = CopyTristanNexus.single_image_nexus(
                data_file[0],
                nexus_file,
                params.input.write_mode,
            )
        elif params.input.experiment_type == "rotation":
            logger.info(
                "Copying metadata for a roation dataset. \n"
                "This means either a multiple images or a multi sequences pump-probe experiment.\n"
            )
            if args.osc_angle:
                logger.info(
                    f"Scan_axis will be a list of values defined by an oscillation angle of {args.osc_angle}."
                )
            elif args.num_bins:
                logger.info(f"Scan_ axis will be a list of {args.num_bins} values.")
            else:
                logger.error(
                    "For a rotation dataset please pass either the oscillation angle or the number of bins."
                )
                sys.exit(
                    "Missing input argument for rotation dataset. Please specify either oscillation angle or number of bins."
                )

            for filename in data_file:
                nxs_img = CopyTristanNexus.multiple_images_nexus(
                    filename,
                    nexus_file,
                    params.input.write_mode,
                    args.osc_angle,
                    args.num_bins,
                )
        logger.info(
            f"Experiment metadata correctly copied from {nexus_file} to {nxs_img}."
        )
    except Exception as err:
        logger.info(f"File {nexus_file} could not be copied.")
        logger.exception(err)


# Define subparsers
subparsers = parser.add_subparsers(
    help="Choose copy methods.",
    required=True,
    dest="sub-command",
)

parser_general = subparsers.add_parser(
    "gen",
    aliases=["copy-file"],
    description=("Copy experiment metadata to a new NeXus file."),
    parents=[full_copy_parser, config_parser],
)
parser_general.set_defaults(func=copy_nexus)

parser_tristan = subparsers.add_parser(
    "tristan",
    aliases=["copy-tristan"],
    description=(
        "Create a new NeXus file for binned images by copying the metadata from the original experiment NeXus file."
    ),
    parents=[tristan_copy_parser, config_parser],
)
parser_tristan.set_defaults(func=copy_tristan_nexus)


def main():
    args = parser.parse_args()
    args.func(args)


# main()



"""
Tools to extract goniometer and detector parameters from GDA JSON files.
"""
from __future__ import annotations

import json
from pathlib import Path
from typing import Dict, Tuple


def read_geometry_from_json(
    axes_geometry: Path | str,
) -> Tuple[Dict, Dict]:
    """
    A function to read the axes information from the GDA-supplied json file.

    Args:
        axes_geometry (Union[Path, str]): JSON file containing the description of the axes and their location.

    Returns:
        Tuple[Dict, Dict]: Returns the updated dictionaries describing goniometer and detector.
    """
    goniometer = {}
    detector = {}

    # Load information from JSON file
    with open(axes_geometry) as f:
        geom = json.load(f)

    coordinate_frame = geom["geometry"]
    print(coordinate_frame)

    # Set up the dictionaries (in theory they are empty even when passed)
    goniometer["axes"] = []
    goniometer["depends"] = []
    goniometer["vectors"] = []
    goniometer["units"] = []
    goniometer["types"] = []

    detector["axes"] = []
    detector["depends"] = []
    detector["vectors"] = []
    detector["units"] = []
    detector["types"] = []

    for v in geom.values():
        if type(v) is dict:
            if v["location"] == "sample":
                goniometer["axes"].append(v["ds_name"])
                goniometer["depends"].append(
                    v["depends_on"].split("_")[1]
                    if "_" in v["depends_on"]
                    else v["depends_on"]
                )
                goniometer["types"].append(v["type"])
                goniometer["units"].append(v["units"])
                [goniometer["vectors"].append(i) for i in v["vector"]]
            elif v["location"] == "detector":
                detector["axes"].append(v["ds_name"])
                detector["depends"].append(v["depends_on"])
                detector["types"].append(v["type"])
                detector["units"].append(v["units"])
                [detector["vectors"].append(i) for i in v["vector"]]

    for j in range(len(goniometer["depends"])):
        if goniometer["depends"][j] in ["x", "y", "z"]:
            goniometer["depends"][j] = "sam_" + goniometer["depends"][j]
    goniometer["offsets"] = len(goniometer["axes"]) * [0, 0, 0]

    return goniometer, detector


def read_detector_params_from_json(
    detector_params: Path | str,
) -> Dict:
    """
    A function to read the detector parameters from the GDA-supplied json file.

    Args:
        detector_params (Union[Path, str]):  JSON file containing the definition of detector parameters.

    Returns:
        Dict: Updated detector dictionary
    """
    detector = {}

    # Load information from JSON file
    with open(detector_params) as f:
        det = json.load(f)

    if "tristan" in det.keys():
        detector["mode"] = "events"
        detector["description"] = det["tristan"]["description"]
        detector["image_size"] = det["tristan"]["data_size_sf"][::-1]
        detector["detector_type"] = det["tristan"]["detector_type"]
        detector["sensor_material"] = (
            "Si"
            if det["tristan"]["sensor_material"] == "Silicon"
            else det["tristan"]["sensor_material"]
        )
        detector["sensor_thickness"] = (
            str(det["tristan"]["sensor_thickness"])
            + det["tristan"]["sensor_thickness_units"]
        )
        detector["pixel_size"] = [
            str(i) + det["tristan"]["pixel_size_units"]
            for i in det["tristan"]["pixel_size_sf"][::-1]
        ]
        detector["fast_axis"] = det["tristan"]["fast_dir"]
        detector["slow_axis"] = det["tristan"]["slow_dir"]
        detector["flatfield_applied"] = False
        detector["pixel_mask_applied"] = False
        # ... and the tristan specifics
        spec = det["detector_specific"]
        detector["software_version"] = spec["software_version"]
        detector["detector_tick"] = (
            str(spec["detector_tick"]) + spec["detector_tick_units"]
        )
        detector["detector_frequency"] = (
            str(spec["detector_frequency"]) + spec["detector_frequency_units"]
        )
        detector["timeslice_rollover"] = spec["timeslice_rollover_bits"]
    elif "eiger" in det.keys():
        detector["mode"] = "images"
        detector["description"] = det["eiger"]["description"]
        detector["detector_type"] = det["eiger"]["type"]
        detector["image_size"] = det["eiger"]["size"]
        detector["sensor_material"] = det["eiger"]["sensor_material"]
        detector["sensor_thickness"] = (
            str(det["eiger"]["thickness"]) + det["eiger"]["thickness_units"]
        )
        detector["pixel_size"] = [
            str(i) + det["eiger"]["pixel_Size_units"]
            for i in det["eiger"]["pixel_size"]
        ]
        detector["fast_axis"] = det["eiger"]["fast_dir"]
        detector["slow_axis"] = det["eiger"]["slow_dir"]
    else:
        print("which detector is this then?")

    return detector



"""
Define a Metafile object to describe the _meta.h5 file and get the necessary information from it.
"""
from __future__ import annotations

import re
from typing import Dict, List, Tuple

import h5py

try:
    # Only for Python version >= 3.8
    from functools import cached_property
except ImportError:
    # Compatibility for earlier Python versions
    import functools

    def cached_property(func):
        @property
        @functools.wraps(func)
        def wrapper_decorator(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper_decorator


__all__ = ["Metafile", "DectrisMetafile", "TristanMetafile"]

tristan_pattern = re.compile(r"ts_qty_module\d{2}")


class Metafile:
    def __init__(self, handle: h5py.File):
        self._handle = handle

    def __getitem__(self, key: str) -> h5py.Group | h5py.Dataset:
        return self._handle[key]

    def __len__(self):
        return len(self._handle)

    def __str__(self):
        return f"File {self._handle.filename} opened in '{self._handle.mode}' mode."

    @cached_property
    def walk(self) -> List[str]:
        obj_list = []
        self._handle.visit(obj_list.append)
        return obj_list

    @cached_property
    def hasMask(self):
        if "mask" in self.walk:
            return True
        return False

    @cached_property
    def hasFlatfield(self):
        if "flatfield" in self.walk:
            return True
        return False


class DectrisMetafile(Metafile):
    """
    Describes a _meta.h5 file for a Dectris Eiger detector.
    """

    def __init__(self, handle: h5py.File):
        super().__init__(handle)

    @cached_property
    def hasDectrisGroup(self) -> bool:
        for k in self._handle.keys():
            if "_dectris" in k and isinstance(self._handle[k], h5py.Group):
                return True
        return False

    @cached_property
    def hasConfig(self) -> bool:
        if "config" in self._handle.keys():
            return True
        return False

    def read_config_dset(self) -> Dict:
        config = eval(self._handle["config"][()])
        return config

    def get_number_of_images(self) -> int:
        if self.hasConfig:
            config = self.read_config_dset()
            if config["nimages"] >= 1 and config["ntrigger"] == 1:
                return config["nimages"]
            elif config["nimages"] == 1 and config["ntrigger"] > 1:
                # For example a "triggered" data collection
                return config["ntrigger"]
        else:
            _loc = [obj for obj in self.walk if "nimages" in obj]
            return self.__getitem__(_loc[0])[0]

    def get_detector_size(self) -> Tuple:
        # NB. returns (fast, slow) but data_size in nxs file shoud be recorded (slow, fast)
        # => det_size[::-1]
        _loc = [obj for obj in self.walk if "pixels_in_detector" in obj]
        det_size = []
        for i in _loc:
            det_size.append(self.__getitem__(i)[0])
        if len(det_size) == 0:
            return None
        return tuple(det_size[::-1])

    def get_pixel_size(self) -> List:
        _loc = [obj for obj in self.walk if "pixel_size" in obj]
        pix = []
        for i in _loc:
            pix.append(self.__getitem__(i)[0])
        if len(pix) == 0:
            return None
        return pix

    def get_beam_center(self) -> List:
        _loc = [obj for obj in self.walk if "beam_center" in obj]
        bc = []
        for i in _loc:
            bc.append(self.__getitem__(i)[0])
        if len(bc) == 0:
            return None
        return bc

    def get_wavelength(self) -> float:
        _loc = [obj for obj in self.walk if "wavelength" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[0]

    def get_detector_distance(self) -> float:
        # Distance in Dectris meta file is in m.
        _loc = [obj for obj in self.walk if "detector_distance" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[0]

    def get_saturation_value(self) -> float:
        _loc = [obj for obj in self.walk if "countrate_correction_count_cutoff" in obj]
        if len(_loc) == 0:
            return None
        return self.__getitem__(_loc[0])[0]

    def get_sensor_information(self) -> Tuple[bytes, float]:
        _loc_material = [obj for obj in self.walk if "sensor_material" in obj]
        _loc_thickness = [obj for obj in self.walk if "sensor_thickness" in obj]
        return (
            self.__getitem__(_loc_material[0])[0],
            self.__getitem__(_loc_thickness[0])[0],
        )

    def find_mask(self) -> Tuple[str, str]:
        if self.hasMask:
            mask_path = [obj for obj in self.walk if obj.lower() == "mask"]
            mask_applied_path = [obj for obj in self.walk if "mask_applied" in obj]
            if len(mask_applied_path) == 0:
                return (mask_path[0], None)
            return (mask_path[0], mask_applied_path[0])
        return (None, None)

    def find_flatfield(self) -> Tuple[str, str]:
        if self.hasFlatfield:
            flatfield_path = [obj for obj in self.walk if obj.lower() == "flatfield"]
            flatfield_applied_path = [
                obj for obj in self.walk if "flatfield_correction_applied" in obj
            ]
            if len(flatfield_applied_path) == 0:
                return (flatfield_path[0], None)
            return (flatfield_path[0], flatfield_applied_path[0])
        return (None, None)

    def find_software_version(self) -> str:
        _loc = [obj for obj in self.walk if "software_version" in obj]
        if len(_loc) == 0:
            return None
        return _loc[0]

    def find_threshold_energy(self) -> str:
        _loc = [obj for obj in self.walk if "threshold_energy" in obj]
        if len(_loc) == 0:
            return None
        return _loc[0]

    def find_bit_depth_readout(self) -> str:
        _loc = [obj for obj in self.walk if "bit_depth_readout" in obj]
        if len(_loc) == 0:
            return None
        return _loc[0]

    def find_detector_number(self) -> str:
        _loc = [obj for obj in self.walk if "detector_number" in obj]
        if len(_loc) == 0:
            return None
        return _loc[0]

    def find_detector_readout_time(self) -> str:
        _loc = [obj for obj in self.walk if "detector_readout_time" in obj]
        if len(_loc) == 0:
            return None
        return _loc[0]


class TristanMetafile(Metafile):
    """
    Describes a _meta.h5 file for a Tristan detector.
    """

    @staticmethod
    def isTristan(filename):
        with h5py.File(filename, "r") as fh:
            res = [k for k in fh.keys() if tristan_pattern.fullmatch(k)]
        if len(res) > 0:
            return True
        return False

    def __init__(self, handle: h5py.File):
        super().__init__(handle)

    def find_number_of_modules(self) -> int:
        n_modules = [k for k in self._handle.keys() if tristan_pattern.fullmatch(k)]
        return len(n_modules)

    def find_software_version(self) -> str:
        _loc = [obj for obj in self.walk if "software_version" in obj]
        if len(_loc) == 0:
            return None
        return _loc[0]

    def find_meta_version(self) -> str:
        _loc = [obj for obj in self.walk if "meta_version" in obj]
        if len(_loc) == 0:
            return None
        return _loc[0]



"""
General tools for blank data writing.
"""
from __future__ import annotations

import logging
import time
from pathlib import Path
from typing import List, Tuple

import h5py
import numpy as np
from hdf5plugin import Bitshuffle
from numpy.typing import ArrayLike

data_logger = logging.getLogger("nexgen.DataWriter")

# Random number generator
rng = np.random.default_rng()

# Eiger specific
eiger_modules = {"1M": (1, 2), "4M": (2, 4), "9M": (3, 6), "16M": (4, 8)}
eiger_mod_size = (512, 1028)
eiger_gap_size = (38, 12)
intra_mod_gap = 2

# Tristan specific
clock_freq = int(6.4e8)
tristan_modules = {"10M": (2, 5), "2M": (1, 2)}
tristan_mod_size = (515, 2069)  # (H, V)
tristan_gap_size = (117, 45)

# Pre-defined chunk size
tristan_chunk = 2097152


# Build-a-detector functions
def build_an_eiger(
    image_size: List | Tuple,
    det_description: str,
    n_modules: Tuple[int, int] = None,
) -> ArrayLike:
    """
    Generate an Eiger-like blank image.

    Args:
        image_size (List | Tuple): Detector size, defines image dimensions as (slow_axis , fast_axis).
        det_description (str): Identifies the type of Eiger detector.
        n_modules (Tuple[int, int], optional): Number of modules in the detector. Defaults to None.

    Returns:
        IM (ArrayLike): Blank image - an array of zeros with an Eiger-like mask.
    """
    for k, v in eiger_modules.items():
        if k in det_description.upper():
            n_modules = v

    IM = np.zeros(image_size, dtype=np.uint16)

    # Horizontal modules
    for i in range(1, n_modules[0] + 1):
        IM[
            :,
            i * eiger_mod_size[1]
            + (i - 1) * eiger_gap_size[1] : i * (eiger_mod_size[1] + eiger_gap_size[1]),
        ] = -1
    # Vertical modules
    for j in range(1, n_modules[1] + 1):
        IM[
            j * eiger_mod_size[0]
            + (j - 1) * eiger_gap_size[0] : j * (eiger_mod_size[0] + eiger_gap_size[0]),
            :,
        ] = -1

    # Intra module gap
    mid = []
    for n in range(n_modules[0]):
        mid.append(
            int(eiger_mod_size[1] / 2) + n * (eiger_mod_size[1] + eiger_gap_size[1])
        )
    for m in mid:
        IM[:, (m - 1) : (m + 1)] = -1

    return IM


def build_a_tristan(
    image_size: List | Tuple,
    det_description: str,
) -> ArrayLike:
    """
    Generate a Tristan-like blank image.

    Args:
        image_size (List | Tuple): Detector size, fefines image dimensions as (slow_axis , fast_axis).
        det_description (str): Identifies the Tristan detector.

    Returns:
        IM (ArrayLike): Blank image - an array of zeros with an Eiger-like mask.
    """
    for k, v in tristan_modules.items():
        if k in det_description.upper():
            n_modules = v

    IM = np.zeros(image_size, dtype=np.uint16)

    # Horizontal modules
    for i in range(1, n_modules[0] + 1):
        IM[
            :,
            i * tristan_mod_size[1]
            + (i - 1)
            * tristan_gap_size[1] : i
            * (tristan_mod_size[1] + tristan_gap_size[1]),
        ] = -1
    # Vertical modules
    for j in range(1, n_modules[1] + 1):
        IM[
            j * tristan_mod_size[0]
            + (j - 1)
            * tristan_gap_size[0] : j
            * (tristan_mod_size[0] + tristan_gap_size[0]),
            :,
        ] = -1

    return IM


def generate_image_files(
    datafiles: List[Path | str],
    image_size: List | Tuple,
    det_description: str,
    tot_num_images: int,
):
    """
    Generate HDF5 files of blank images.

    Args:
        datafiles (List[Path | str]): List of HDF5 files to be written.
        image_size (List | Tuple): Image dimensions as (slow_axis, fast_axis).
        det_description (str): Type of detector. The string should include the number of modules.
        tot_num_images (int): Total number of images to be written across the files.

    Raises:
        ValueError: If the number of files requested and the total number of images to write don't match.
    """
    # Write some blank data in the shape of a detector
    if "eiger" in det_description.lower():
        img = build_an_eiger(image_size, det_description)
    elif "tristan" in det_description.lower():
        img = build_a_tristan(image_size, det_description)
    else:
        # Do nothing for now, just add zeros
        img = np.zeros(image_size, dtype=np.uint16)

    # Determine single dataset shape: (num, *img_size), where max(num)=1000.
    dset_shape = (tot_num_images // 1000) * [1000] + [tot_num_images % 1000]
    # if tot_num_images <= 1000:
    #    dset_shape = [tot_num_images]
    # elif tot_num_images % 1000 == 0:
    #    dset_shape = (tot_num_images // 1000) * [1000]
    # else:
    #    dset_shape = (tot_num_images // 1000) * [1000] + [tot_num_images % 1000]

    # Just a quick check
    if len(dset_shape) != len(datafiles):
        raise ValueError(
            "Impossible to write blank images to file: number of files desn't match the dataset shape."
        )

    # Start writing files
    for filename, sh0 in zip(datafiles, dset_shape):
        data_logger.info(f"Writing {filename} ...")
        tic = time.process_time()
        with h5py.File(filename, "w") as fh:
            dset = fh.create_dataset(
                "data",
                shape=(sh0, *image_size),
                dtype=np.uint16,
                chunks=(1, *image_size),
                **Bitshuffle(),
            )
            # Use direct chunk write
            dset[0, :, :] = img
            f, ch = dset.id.read_direct_chunk((0, 0, 0))
            for j in range(1, sh0):
                dset.id.write_direct_chunk((j, 0, 0), ch, f)
        toc = time.process_time()
        data_logger.info(f"Writing {sh0} images took {toc - tic:.2f} s.")


# Event list generator
# TODO Better than before, but this is still pretty slow.
def pseudo_event_list(
    x_lim: Tuple[int, int | None],
    y_lim: Tuple[int, int | None],
    exp_time: float,
) -> Tuple[List, List]:
    """
    Generate a pseudo-events list with positions and timestamps.

    Args:
        x_lim (Tuple[int, Union[int, None]]): Minimum and maximum position along the fast axis.
        y_lim (Tuple[int, Union[int, None]]): Minimum and maximum position along the slow axis.
        exp_time (float): Total exposure time, in seconds.

    Returns:
        pos_list, time_list (Tuple[List, List]): Lists of pseudo-event positions and relative timestamps.
    """
    pos_list = []
    time_list = []

    for _ in range(tristan_chunk):
        dist = rng.uniform(0, 1)
        x = (
            np.random.randint(x_lim[0], x_lim[1])
            if len(x_lim) > 1
            else np.random.randint(x_lim[0])
        )
        y = (
            np.random.randint(y_lim[0], y_lim[1])
            if len(y_lim) > 1
            else np.random.randint(y_lim[0])
        )
        loc = np.uint32(x * np.uint32(0x2000) + y)
        pos_list.append(loc)
        t = np.uint64((exp_time + dist) * clock_freq)
        time_list.append(t)

    return pos_list, time_list


def generate_event_files(
    datafiles: List[Path | str],
    num_chunks: int,
    det_description: str,
    exp_time: float,
):
    """
    Generate HDF5 files of pseudo events.

    Args:
        datafiles (List[Union[Path, str]]): List of HDF5 files to be written.
        num_chunks (int): Chunks of events to be written per file.
        det_description (str): Type of detector. The string should include the number of modules.
        exp_time (float): Total exposure time, in seconds.
    """
    # A bunch of things to be done here first ...
    # Get number of modules in the Tristan detector
    for k, v in tristan_modules.items():
        if k in det_description.upper():
            n_modules = v

    # Some blank cues
    blank_cues = np.zeros(tristan_chunk, dtype=np.uint16)

    # TODO FIXME speed this up!
    data_logger.info(
        f"Start generating one chunk of pseudo events for {n_modules} modules of {det_description}"
    )
    t0 = time.process_time()
    EV_dict = {}
    for i in range(n_modules[0]):
        for j in range(n_modules[1]):
            I = (
                i * (tristan_mod_size[1] + tristan_gap_size[1]),
                (i + 1) * tristan_mod_size[1] + i * tristan_gap_size[1],
            )
            J = (
                j * (tristan_mod_size[0] + tristan_gap_size[0]),
                (j + 1) * tristan_mod_size[0] + j * tristan_gap_size[0],
            )
            EV_dict[(i, j)] = pseudo_event_list(I, J, exp_time)
    t1 = time.process_time()
    data_logger.info(f"Time taken to generate pseudo-event list: {t1-t0:.2f} s.")

    # Find total number of events to be written to file
    num_events = tristan_chunk * num_chunks

    # Start writing files
    for filename, K in zip(datafiles, EV_dict.keys()):
        data_logger.info(f"Writing {filename} ...")
        tic = time.process_time()
        with h5py.File(filename, "w") as fh:
            fh.create_dataset("cue_id", data=blank_cues, **Bitshuffle())
            fh.create_dataset("cue_timestamp_zero", data=blank_cues, **Bitshuffle())
            ev_id = fh.create_dataset(
                "event_id",
                shape=(num_events,),
                dtype=np.uint32,
                chunks=(tristan_chunk,),
                **Bitshuffle(),
            )
            ev_t = fh.create_dataset(
                "event_timestamp_zero",
                shape=(num_events,),
                dtype=np.uint64,
                chunks=(tristan_chunk,),
                **Bitshuffle(),
            )
            ev_en = fh.create_dataset(
                "event_energy",
                shape=(num_events,),
                dtype=np.uint32,
                chunks=(tristan_chunk,),
                **Bitshuffle(),
            )

            # Use direct chunk write
            ev_id[:tristan_chunk] = EV_dict[K][0]
            ev_t[:tristan_chunk] = EV_dict[K][1]
            ev_en[:tristan_chunk] = blank_cues
            f_id, ch_id = ev_id.id.read_direct_chunk((0,))
            f_t, ch_t = ev_t.id.read_direct_chunk((0,))
            f_en, ch_en = ev_en.id.read_direct_chunk((0,))
            for h in range(1, num_chunks):
                ev_id.id.write_direct_chunk((h * tristan_chunk,), ch_id, f_id)
                ev_t.id.write_direct_chunk((h * tristan_chunk,), ch_t, f_t)
                ev_en.id.write_direct_chunk((h * tristan_chunk,), ch_en, f_en)
        toc = time.process_time()
        data_logger.info(f"Writing {num_events} events took {toc - tic:.2f} s.")



"""
Utilities for writing new NeXus format files.
"""

from __future__ import annotations

import math
from pathlib import Path
from typing import Dict, List, Tuple

# import h5py
import numpy as np
from hdf5plugin import Bitshuffle
from numpy.typing import ArrayLike

# from h5py import AttributeManager
from scanspec.core import Path as ScanPath
from scanspec.specs import Line

import h5py  # isort: skip
from h5py import AttributeManager  # isort: skip


def create_attributes(nxs_obj: h5py.Group | h5py.Dataset, names: Tuple, values: Tuple):
    """
    Create or overwrite attributes with additional metadata information.

    Args:
        nxs_obj (h5py.Group | h5py.Dataset): NeXus object to which the attributes should be attached.
        names (Tuple): The names of the new attributes.
        values (Tuple): The attribute values asociated to the names.
    """
    for n, v in zip(names, values):
        if type(v) is str:
            # If a string, convert to numpy.string_
            v = np.string_(v)
        AttributeManager.create(nxs_obj, name=n, data=v)


def set_dependency(dep_info: str, path: str = None):
    """
    Define value for "depends_on" attribute.
    If the attribute points to the head of the dependency chain, simply pass "." for dep_info.

    Args:
        dep_info (str): The name of the transformation upon which the current one depends on.
        path (str): Where the transformation is. Set to None, if passed it points to location in the NeXus tree.
    Returns:
        The value to be passed to the attribute "depends_on"
    """
    if dep_info == ".":
        return np.string_(".")
    if path:
        if path.endswith("/") is False:
            path += "/"
        return np.string_(path + dep_info)
    else:
        return np.string_(dep_info)


def find_osc_axis(
    axes_names: List,
    axes_starts: List,
    axes_ends: List,
    axes_types: List,
    default: str = "omega",
) -> str:
    """
    Identify the rotation scan_axis.

    This function identifies the scan axis from the list passed as argument.
    The scan axis is the one where start and end value are not the same.
    If there is only one rotation axis, that is the one returned.
    In the case scan axis cannot be identified, a default value is arbitrarily assigned.

    Args:
        axes_names (List): List of names associated to goniometer axes.
        axes_starts (List): List of start values.
        axes_ends (List): List of end values.
        axes_types (List): List of axes types, useful to identify only the rotation axes.
        default (str, optional): String to deafult to in case scan axis is not found. Defaults to "omega".

    Raises:
        ValueError: If no axes have been passed.
        ValueError: If more than one rotation axis seems to move.

    Returns:
        scan_axis (str): String identifying the rotation scan axis.
    """
    # This assumes that at least one rotation axis is always passed.
    # Assuming all list are of the same length ...
    if len(axes_names) == 0:
        raise ValueError(
            "Impossible to determine translation scan. No axes passed to find_osc_axis function. Please make sure at least one value is passed."
        )
    # assert len(axes_names) > 0, "Please pass at least one axis."
    # Look only for rotation axes
    rot_idx = [i for i in range(len(axes_types)) if axes_types[i] == "rotation"]
    axes_names = [axes_names[j] for j in rot_idx]
    axes_starts = [axes_starts[j] for j in rot_idx]
    axes_ends = [axes_ends[j] for j in rot_idx]

    if len(axes_names) == 1:
        scan_axis = axes_names[0]
    else:
        idx = [(i != j) for i, j in zip(axes_starts, axes_ends)]
        if idx.count(True) == 0:
            # just in case ...
            scan_axis = default
        elif idx.count(True) == 1:
            scan_axis = axes_names[idx.index(True)]
        else:
            raise ValueError("Unable to correctly identify the rotation scan axis.")
    return scan_axis


def find_grid_scan_axes(
    axes_names: List,
    axes_starts: List,
    axes_ends: List,
    axes_types: List,
) -> List[str]:
    """
    Identify the scan axes for a linear/grid scan.

    Args:
        axes_names (List): List of names associated to goniometer axes.
        axes_starts (List): List of start values.
        axes_ends (List): List of end values.
        axes_types (List): List of axes types, useful to identify only the translation axes.

    Raises:
        ValueError: If no axes have been passed.

    Returns:
        scan_axis (List[str]): List of strings identifying the linear/grid scan axes. If no axes are identified, it will return an empty list.
    """
    if len(axes_names) == 0:
        raise ValueError(
            "Impossible to determine translation scan. No axes passed to find_grid_scan_axes function. Please make sure at least one value is passed."
        )

    # Look only at translation axes
    grid_idx = [i for i in range(len(axes_names)) if axes_types[i] == "translation"]
    axes_names = [axes_names[j] for j in grid_idx]
    axes_starts = [axes_starts[j] for j in grid_idx]
    axes_ends = [axes_ends[j] for j in grid_idx]

    scan_axis = []
    for n, ax in enumerate(axes_names):
        if axes_starts[n] != axes_ends[n]:
            scan_axis.append(ax)
    return scan_axis


def calculate_scan_range(
    axes_names: List,
    axes_starts: List,
    axes_ends: List,
    axes_increments: List = None,
    n_images: Tuple | int = None,
    snaked: bool = True,
    rotation: bool = False,
) -> Dict[str, ArrayLike]:
    """
    Calculate the scan range for a linear/grid scan or a rotation scan from the number of images to be written.
    If the number of images is not provided, it can be calculated from the increment value of the axis (these values are mutually exclusive).
    When dealing with a rotation axis, if there are multiple images but no rotation scan, return axis_start repeated n_images times.

    Args:
        axes_names (List): List of names for the axes involved in the scan.
        axes_starts (List): List of axis positions at the beginning of the scan.
        axes_ends (List): List of axis positions at the end of the scan.
        axes_increments (List, optional): List of ranges through which the axes move each frame. Mostly used for rotation scans. Defaults to None.
        n_images (Tuple | int, optional): Number of images to be written. If writing a 2D scan, it should be a (nx, ny) tuple, \
                                        where tot_n_img=nx*ny, any int value is at this time ignored. Defaults to None.
        snaked (bool): If True, scanspec will "draw" a snaked grid. Defaults to True.
        rotation (bool): Tell the function to calculate a rotation scan. Defaults to False.

    Raises:
        TypeError: If the input axes are not lists.
        ValueError: When an empty axes names list has been passed.
        ValueError: When both axes_increments and n_images have been passed. The two values are mutually exclusive.
        ValueError: When neither axes_increments not n_images have been passed.
        ValueError: For a grid scan, if axes_increments is None, n_images must be a tuple of len=2 to be sure to accurately calculate the scan points.

    Returns:
        Dict[str, ArrayLike]: A dictionary of ("axis_name": axis_range) key-value pairs.
    """
    if type(axes_names) != list or type(axes_starts) != list or type(axes_ends) != list:
        raise TypeError("Input values for axes must be passed as lists.")

    if len(axes_names) == 0:
        raise ValueError("No axes have been passed, impossible to determine scan.")

    if n_images and axes_increments:
        raise ValueError(
            "The axes_increments and n_images arguments are mutually exclusive. Please pass just one of those."
            "For a 2D scan it is recommended that n_images is passed."
        )
    elif not n_images and not axes_increments:
        raise ValueError(
            "Impossible to calculate scan points, please pass either the axes increment values or the number of scan points (n_images) per axis."
            "For a 2D scan it is recommended that n_images is passed."
        )

    if len(axes_names) == 1 and rotation is True:
        if not n_images:
            n_images = round(abs(axes_starts[0] - axes_ends[0]) / axes_increments[0])

        if axes_starts[0] != axes_ends[0] and axes_increments:
            if axes_starts[0] > axes_ends[0]:
                # Account for reverse rotation.
                axes_ends[0] = axes_ends[0] + axes_increments[0]
            else:
                axes_ends[0] = axes_ends[0] - axes_increments[0]
        elif axes_starts[0] != axes_ends[0] and not axes_increments:
            inc = (axes_ends[0] - axes_starts[0]) / n_images
            axes_ends[0] = axes_ends[0] - inc

        spec = Line(axes_names[0], axes_starts[0], axes_ends[0], n_images)
        scan_path = ScanPath(spec.calculate())

    elif len(axes_names) == 1 and rotation is False:
        if not n_images:
            # FIXME This calculation still gives the wrong increment between scan points.
            n_images = (
                round(abs(axes_starts[0] - axes_ends[0]) / axes_increments[0]) + 1
            )
        elif type(n_images) is tuple and len(n_images) == 1:
            # This is mostly a double paranoid check
            n_images = n_images[0]

        spec = Line(axes_names[0], axes_starts[0], axes_ends[0], n_images)
        scan_path = ScanPath(spec.calculate())

    else:
        if not n_images:
            # FIXME This calculation still gives the wrong increment between scan points.
            n_images0 = (
                round(abs(axes_starts[0] - axes_ends[0]) / axes_increments[0]) + 1
            )
            n_images1 = (
                round(abs(axes_starts[1] - axes_ends[1]) / axes_increments[1]) + 1
            )
        elif len(n_images) == 1:
            raise ValueError(
                "Impossible to correctly calculate scan points from just the total number of images."
                "Please either pass a tuple with the number of scan point per axis or the axes increments."
            )
        else:
            n_images0 = n_images[0]
            n_images1 = n_images[1]

        if snaked is True:
            spec = Line(axes_names[0], axes_starts[0], axes_ends[0], n_images0) * ~Line(
                axes_names[1], axes_starts[1], axes_ends[1], n_images1
            )
        else:
            spec = Line(axes_names[0], axes_starts[0], axes_ends[0], n_images0) * Line(
                axes_names[1], axes_starts[1], axes_ends[1], n_images1
            )
        scan_path = ScanPath(spec.calculate())

    return scan_path.consume().midpoints


def calculate_origin(
    beam_center_fs: List | Tuple,
    fs_pixel_size: List | Tuple,
    fast_axis_vector: Tuple,
    slow_axis_vector: Tuple,
    mode: str = "1",
) -> Tuple[List, float]:
    """
    Calculate the offset of the detector.

    This function returns the detector origin array, which is saved as the vector attribute of the module_offset field.
    The value to set the module_offset to is also returned: the magnitude of the displacement if the vector is normalized, 1.0 otherwise
    Assumes that fast and slow axis vectors have already been converted to mcstas if needed.

    Args:
        beam_center_fs (List | Tuple): Beam center position in fast and slow direction.
        fs_pixel_size (List | Tuple): Pixel size in fast and slow direction, in m.
        fast_axis_vector (Tuple): Fast axis vector.
        slow_axis_vector (Tuple): Slow axis vector.
        mode (str, optional): Decide how origin should be calculated.
                            If set to "1" the displacement vector is un-normalized and the offset value set to 1.0.
                            If set to "2" the displacement is normalized and the offset value is set to the magnitude of the displacement.
                            Defaults to "1".

    Returns:
        det_origin (List): Displacement of beam center, vector attribute of module_offset.
        offset_val (float): Value to assign to module_offset, depending whether det_origin is normalized or not.
    """
    # what was calculate module_offset
    x_scaled = beam_center_fs[0] * fs_pixel_size[0]
    y_scaled = beam_center_fs[1] * fs_pixel_size[1]
    # Detector origin
    det_origin = x_scaled * np.array(fast_axis_vector) + y_scaled * np.array(
        slow_axis_vector
    )
    det_origin = list(-det_origin)
    if mode == "1":
        offset_val = 1.0
    else:
        offset_val = math.hypot(*det_origin[:-1])
    return det_origin, offset_val


def find_number_of_images(datafile_list: List[Path], entry_key: str = "data") -> int:
    """
    Calculate total number of images when there's more than one input HDF5 file.

    Args:
        datafile_list (List[Path]): List of paths to the input image files.
        entry_key (str):    Key for the location of the images inside the data files. Defaults to "data".

    Returns:
        num_images (int): Total number of images.
    """
    num_images = 0
    for filename in datafile_list:
        with h5py.File(filename, "r") as f:
            num_images += f[entry_key].shape[0]
    return int(num_images)


def set_instrument_name(source: Dict, facility_id: str = "DIAMOND") -> str:
    """Set the instrument name from the details saved in source.

    If source type is not defined in the input dictionary, the function will assume it is a Synchrotron.
    If the facility_id is defined inside the source dictionary, that is the value that will be used.
    Naming tries to follow the recommended convention for NXmx:
    https://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Items/_diffrn_source.type.html

    Args:
        source (Dict): Dictionary containing the facility information. \
            Minimum required keys: name, beamline_name
        facility_id (str, optional): Facility identifier. Defaults to "DIAMOND".

    Returns:
        name (str): The name to write under '/entry/instrument/name'
    """
    if "facility_id" in source.keys() and source["facility_id"] is not None:
        facility_id = source["facility_id"]

    if "type" not in source.keys() or source["type"] is None:
        return f"{facility_id} BEAMLINE {source['beamline_name']}"

    if "SYNCHROTRON" in source["type"].upper():
        name = f"{facility_id} BEAMLINE {source['beamline_name']}"
    else:
        name = f"{facility_id} {source['beamline_name']}"
    return name


# Copy and compress a dataset inside a specified NXclass
def write_compressed_copy(
    nxgroup: h5py.Group,
    dset_name: str,
    data: ArrayLike = None,
    filename: Path | str = None,
    dset_key: str = None,
    block_size: int = 0,
):
    """
    Write a compressed copy of some dataset in the desired HDF5 group, using the Bitshuffle filter with lz4 compression.
    The main application for this function in nexgen is to write a compressed copy of a pixel mask or a flatfield file/dataset \
    directly into the NXdetector group of a NXmx NeXus file.
    The data and filename arguments are mutually exclusive as only one of them can be used as input.
    If a filename is passed, it is also required to pass the key for the relevant dataset to be copied. Failure to do so will result \
    in nothing being written to the NeXus file.

    Args:
        nxgroup (h5py.Group): Handle to HDF5 group.
        dset_name (str): Name of the new dataset to be written.
        data (ArrayLike, optional): Dataset to be compressed. Defaults to None.
        filename (Path | str, optional): Filename containing the dataset to be compressed into the NeXus file. Defaults to None.
        dset_key (str, optional): Dataset name inside the passed file. Defaults to None.
        block_size (int, optional): Number of elements per block, it needs to be divisible by 8. Defaults to 0.

    Raises:
        ValueError: If both a dataset and a filename have been passed to the function.
    """
    from .NXclassWriters import NXclass_logger

    if data is not None and filename is not None:
        raise ValueError(
            "The dset and filename arguments are mutually exclusive."
            "Please pass only the one from which the data should be copied."
        )
    if filename and not dset_key:
        NXclass_logger.warning(
            f"Missing key to find the dataset to be copied inside {filename}. {dset_name} will not be written into the NeXus file."
        )
        return

    if filename:
        with h5py.File(filename, "r") as fh:
            data = fh[dset_key][()]

    nxgroup.create_dataset(
        dset_name, data=data, **Bitshuffle(nelems=block_size, cname="lz4")  # lz4=True)
    )
    NXclass_logger.info(
        f"A compressed copy of the {dset_name} has been written into the NeXus file."
    )



"""
Create a NeXus file for serial crystallography datasets collected on Tristan10M detector either on I19-2 or I24 beamlines.
"""

import logging
from collections import namedtuple
from pathlib import Path

import h5py

from .. import get_iso_timestamp, get_nexus_filename, log
from ..nxs_write.NexusWriter import call_writers
from ..nxs_write.NXclassWriters import write_NXdatetime, write_NXentry, write_NXnote
from . import source
from .I19_2_params import tristan10M_module as module
from .I19_2_params import tristan10M_params as detector

__all__ = ["ssx_tristan_writer"]

# Define a logger object and a formatter
logger = logging.getLogger("nexgen.SSX_Tristan")

ssx_tr_collect = namedtuple(
    "ssx_collect",
    [
        "exposure_time",
        "detector_distance",
        "beam_center",
        "transmission",
        "wavelength",
        "start_time",
        "stop_time",
        "chipmap",
        "chip_info",
    ],
)

ssx_tr_collect.__doc__ = (
    """Parameters that define a serial collection using a Tristan detector."""
)

# Define coordinate frame
coordinate_frame = "mcstas"

# Initialize dictionaries
beam = {}
attenuator = {}


def ssx_tristan_writer(
    visitpath: Path | str,
    filename: str,
    beamline: str,
    **ssx_params,
):
    """
    Gather all parameters from the beamline and call the NeXus writers.

    Args:
        visitpath (Path | str): Path to colection directory.
        filename (str): Root of the filename.
        beamline (str): Beamline on which the experiment is being run.

    Keyword Args:
        exp_time (float): Exposure time, in s.
        det_dist (float): Distance between sample and detector, in mm.
        beam_center (List[float, float]): Beam center position, in pixels.
        transmission (float): Attenuator transmission, in %.
        wavelength (float): Wavelength of incident beam, in A.
        flux (float): Total flux.
        start_time (datetime): Experiment start time.
        stop_time (datetime): Experiment end time.
        chip_info (Dict): For a grid scan, dictionary containing basic chip information.
            At least it should contain: x/y_start, x/y number of blocks and block size, x/y number of steps and number of exposures.
        chipmap (Path | str): Path to the chipmap file corresponding to the experiment,
            or 'fullchip' indicating that the whole chip is being scanned.
    """
    # Get info from the beamline
    SSX_TR = ssx_tr_collect(
        exposure_time=float(ssx_params["exp_time"]),
        detector_distance=float(ssx_params["det_dist"]),
        beam_center=ssx_params["beam_center"],
        transmission=float(ssx_params["transmission"]),
        wavelength=float(ssx_params["wavelength"]),
        start_time=ssx_params["start_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if ssx_params["start_time"]
        else None,  # This should be datetiem type
        stop_time=ssx_params["stop_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if ssx_params["stop_time"]
        else None,  # idem.
        chipmap=ssx_params["chipmap"] if ssx_params["chipmap"] else None,
        chip_info=ssx_params["chip_info"] if ssx_params["chip_info"] else None,
    )

    visitpath = Path(visitpath).expanduser().resolve()
    filename = ssx_params["filename"]

    logfile = SSX_TR.visitpath / f"{beamline}_TristanSSX_nxs_writer.log"
    # Configure logging
    log.config(logfile.as_posix())

    logger.info(
        f"Start NeXus File Writer for time-resolved SSX on beamline {beamline} at DLS."
    )

    logger.info(f"Current collection directory: {visitpath}")
    # Find metafile in directory and get info from it
    try:
        metafile = [
            f for f in visitpath.iterdir() if filename + "_meta" in f.as_posix()
        ][0]
        logger.info(f"Found {metafile} in directory.")
    except IndexError as err:
        logger.exception(err)
        logger.error(
            "Missing metadata, something might be wrong with this collection."
            "Unable to write NeXus file at this time. Please try using command line tool."
        )
        raise

    # Add some information to logger
    logger.info("Creating a NeXus file for %s ..." % metafile.name)
    # Get NeXus filename
    master_file = get_nexus_filename(metafile)
    logger.info("NeXus file will be saved as %s" % master_file)

    # Get parameters depending on beamline
    logger.info(f"DLS Beamline: {beamline.upper()}.")
    if "I19" in beamline.upper():
        from .I19_2_params import goniometer_axes as goniometer

        source["beamline_name"] = beamline.upper()
        beam["flux"] = None
    elif "I24" in beamline.upper():
        from .I24_Eiger_params import goniometer_axes as goniometer

        source["beamline_name"] = beamline.upper()
        beam["flux"] = ssx_params["flux"] if "flux" in ssx_params.keys() else None
    else:
        raise ValueError(
            "Unknown beamline for SSX collections with Tristan 10M detector."
            "Beamlines currently enabled for the writer: I24, I19-2."
        )

    # Add to dictionaries
    # Detector
    # If location is I24, two_theta is not present
    detector["starts"] = (
        [0.0, SSX_TR.detector_distance]
        if "I19" in beamline.upper()
        else [SSX_TR.detector_distance]
    )
    if "I24" in beamline.upper():
        detector["axes"] = ["det_z"]
        detector["types"] = ["translation"]
        detector["units"] = ["mm"]
        detector["depends"] = ["."]
        detector["vectors"] = [0, 0, 1]
        detector["increments"] = [0.0]

    detector["exposure_time"] = SSX_TR.exposure_time
    detector["beam_center"] = SSX_TR.beam_center

    # Attenuator
    attenuator["transmission"] = SSX_TR.transmission

    # Beam
    beam["wavelength"] = SSX_TR.wavelength

    l = len(goniometer["axes"])
    goniometer["starts"] = goniometer["ends"] = goniometer["increments"] = l * [0.0]

    # Get rotation scan range array and axis
    osc_axis = "phi" if "I19" in SSX_TR.location else "omega"
    osc_range = (0.0, 0.0)

    OSC = {osc_axis: osc_range}

    # Get timestamps in the correct format
    timestamps = (
        get_iso_timestamp(SSX_TR.start_time),
        get_iso_timestamp(SSX_TR.stop_time),
    )

    logger.info("--- COLLECTION SUMMARY ---")
    logger.info("Source information")
    logger.info(f"Facility: {source['name']} - {source['type']}.")
    logger.info(f"Beamline: {source['beamline_name']}")

    logger.info(f"Incident beam wavelength: {beam['wavelength']}")
    logger.info(f"Attenuation: {attenuator['transmission']}")

    logger.info("Goniometer information")
    for j in range(len(goniometer["axes"])):
        logger.info(
            f"Goniometer axis: {goniometer['axes'][j]} => {goniometer['starts'][j]}, {goniometer['types'][j]} on {goniometer['depends'][j]}"
        )
    logger.info("Detector information")
    logger.info(f"{detector['description']}")
    logger.info(
        f"Sensor made of {detector['sensor_material']} x {detector['sensor_thickness']}"
    )
    logger.info(
        f"Detector is a {detector['image_size'][::-1]} array of {detector['pixel_size']} pixels"
    )
    for k in range(len(detector["axes"])):
        logger.info(
            f"Detector axis: {detector['axes'][k]} => {detector['starts'][k]}, {detector['types'][k]} on {detector['depends'][k]}"
        )

    logger.info(f"Recorded beam center is: {detector['beam_center']}.")

    logger.info(f"Timestamps recorded: {timestamps}.")

    try:
        with h5py.File(master_file, "x") as nxsfile:
            write_NXentry(nxsfile)

            if timestamps[0]:
                write_NXdatetime(nxsfile, (timestamps[0], None))

            call_writers(
                nxsfile,
                [metafile],
                coordinate_frame,
                (
                    detector["mode"],
                    None,
                ),  # since it's event mode, don't need event number or chunk number here
                goniometer,
                detector,
                module,
                source,
                beam,
                attenuator,
                OSC,
                transl_scan=None,
                metafile=metafile,  # Since there are no links, this could also be None
                link_list=None,
            )

            # Save chipmap (list of city blocks)
            if SSX_TR.chipmap:
                # Assuming 8x8 fast chip
                from .SSX_chip import read_chip_map

                chip = read_chip_map(SSX_TR.chipmap, 8, 8)
                mapping = {"chipmap": str(chip)}
                logger.info(
                    f"Chipmap read from {SSX_TR.chipmap}, saving in '/entry/source/notes/chipmap'."
                )
                write_NXnote(nxsfile, "/entry/source/notes/", mapping)
                # To read this: eval(dset[()])
            # Save chip info if passed. If not save I24-like chip info plus warning message
            logger.info("Save chip information in /entry/source/notes/chip")
            if SSX_TR.chip_info:
                # Make chip info more readable
                chip_info = {k: v[1] for k, v in SSX_TR.chip_info.items()}
                chipdef = {"chip": str(chip_info)}
                write_NXnote(nxsfile, "/entry/source/notes/", chipdef)
            else:
                logger.warning(
                    f"Dictionary containing chip info was not passed to the writer."
                    "The following values will be written as default: "
                    "x/y_num_blocks = 8 \n x/y_block_size = 3.175 \n x/y_num_steps = 20 \n x/y_step_size = 0.125"
                )
                from .SSX_chip import CHIP_DICT_DEFAULT as chip_info

                chipdef = {"chip": str(chip_info)}
                write_NXnote(nxsfile, "/entry/source/notes", chipdef)

            if timestamps[1]:
                write_NXdatetime(nxsfile, (None, timestamps[1]))
            logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )
        raise



"""
IO tool to gather beamline and collection information from xml file.
"""

import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Dict, Tuple, Union


class ExtendedRequestIO:
    """Define an ExtendedRequest object which in GDA gathers all the information regarding beamline and collection into an xml file."""

    def __init__(self, xmlfile: Union[Path, str]):
        self.tree = ET.parse(xmlfile)
        self.root = self.tree.getroot()

    def getCollectionInfo(self) -> Tuple[str, str, str]:
        directory = self.root.find(".//directory").text
        prefix = self.root.find(".//prefix").text
        run_number = self.root.find(".//runNumber").text
        return directory, prefix, run_number

    def getOscillationSequence(self) -> Dict[str, float]:
        osc_seq_node = self.root.find(".//oscillation_sequence")
        osc_seq = {
            "start": float(osc_seq_node.find(".//start").text),
            "range": float(osc_seq_node.find(".//range").text),
            "number_of_images": float(osc_seq_node.find(".//number_of_images").text),
            "exposure_time": float(osc_seq_node.find(".//exposure_time").text),
            "number_of_passes": float(osc_seq_node.find(".//number_of_passes").text),
        }
        return osc_seq

    def getAxisChoice(self) -> str:
        ax = self.root.find(".//axisChoice").text
        return ax.lower()

    def getOtherAxis(self) -> float:
        val = float(self.root.find(".//otherAxis").text)
        return val

    def getKappa(self) -> float:
        return float(self.root.find(".//kappa").text)

    def getChi(self) -> float:
        return float(self.root.find(".//chi").text)

    def getTwoTheta(self) -> float:
        return float(self.root.find(".//twoTheta").text)

    def getSampleDetectorDistance(self) -> float:
        return float(self.root.find(".//sampleDetectorDistanceInMM").text)

    def getTransmission(self) -> float:
        return float(self.root.find(".//transmissionInPerCent").text)

    def getResolution(self) -> float:
        return float(self.root.find(".//resolution").text)
